"""Module for common tools used throughout the project
"""

from functools import wraps
from logging import Logger, getLogger
from typing import Any, Callable, TypeVar, cast

F = TypeVar("F", bound=Callable[..., Any])


def log_details(logger: Logger, func: F, *args: Any, **kwargs: Any) -> F:
    """Logs details of a function call

    Args:
        logger (logging.Logger)
        func (F)

    Returns:
        F
    """
    logger = logger or getLogger(func.__module__)
    logger.debug(
        "%s called\n\targs: %s\n\tkwargs: %s",
        func.__name__,
        "\n\t\t".join([str(s) for s in args]),
        "\n\t\t".join(
            [f"{key}: {value}" for key, value in kwargs.items()],
        ),
    )
    return_ = func(*args, **kwargs)
    logger.debug("Returning: %s", return_)
    return cast(F, return_)


def log_call(func: F) -> F:
    """Used to log calls to the function provided"""

    @wraps(func)
    def wrapper(*args, **kwargs):  # type: ignore
        logger = getLogger(func.__module__)
        return log_details(logger, func, *args, **kwargs)

    return cast(F, wrapper)


def untested(func: F) -> F:
    """Used to log a warning that the decorated function has not been tested"""

    @wraps(func)
    def wrapper(*args, **kwargs):  # type: ignore
        logger = getLogger(func.__module__)
        logger.warning("Calling untested function: %s", func.__name__)
        return log_details(logger, func, *args, **kwargs)

    return cast(F, wrapper)



"""Classes related to card endpoints
"""

from __future__ import annotations

from logging import getLogger
from typing import Any, ClassVar, Optional

from metabase_tools.endpoints.generic_endpoint import Endpoint
from metabase_tools.models.generic_model import MissingParam
from metabase_tools.models.user_model import UserItem
from metabase_tools.utils.logging_utils import log_call

logger = getLogger(__name__)


class Users(Endpoint[UserItem]):
    """Card related endpoint methods"""

    _BASE_EP: ClassVar[str] = "/user"
    _STD_OBJ: ClassVar[type] = UserItem
    _required_params: ClassVar[list[str]] = [
        "first_name",
        "last_name",
        "email",
        "password",
    ]

    @log_call
    def get(self, targets: Optional[list[int]] = None) -> list[UserItem]:
        """Fetch list of users

        Args:
            targets (list[int], optional): If provided, the list of users to fetch

        Returns:
            list[UserItem]
        """
        return super().get(targets=targets)

    def _make_create(self, **kwargs: Any) -> UserItem:
        """Makes create request

        Args:
            self (UserItem)

        Returns:
            UserItem: self
        """
        return super()._make_create(**kwargs)

    @log_call
    def create(
        self,
        first_name: str | MissingParam = MissingParam(),
        last_name: str | MissingParam = MissingParam(),
        email: str | MissingParam = MissingParam(),
        password: str | MissingParam = MissingParam(),
        group_ids: Optional[list[int] | MissingParam] = MissingParam(),
        login_attributes: Optional[str | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> UserItem:
        """_summary_

        Args:
            first_name (str, optional)
            last_name (str, optional)
            email (str, optional)
            password (str, optional)
            group_ids (list[int], optional)
            login_attributes (str, optional)

        Returns:
            UserItem: _description_
        """
        return super()._make_create(
            first_name=first_name,
            last_name=last_name,
            email=email,
            password=password,
            group_ids=group_ids,
            login_attributes=login_attributes,
            **kwargs,
        )

    @log_call
    def search(
        self,
        search_params: list[dict[str, Any]],
        search_list: Optional[list[UserItem]] = None,
    ) -> list[UserItem]:
        """Method to search a list of users meeting a list of parameters

        Args:
            search_params (list[dict]): Each dict contains search criteria and returns\
                 1 result
            search_list (list[UserItem], optional): Provide to search an existing \
                list, by default pulls from API

        Returns:
            list[UserItem]: List of users of the relevant type
        """
        return super().search(search_params=search_params, search_list=search_list)

    @log_call
    def current(self) -> UserItem:
        """Current user details

        Raises:
            RequestFailure: Invalid response from API

        Returns:
            User: Current user details
        """
        result = self._adapter.get(endpoint="/user/current")
        if isinstance(result, dict):
            return UserItem(**result)
        raise TypeError(f"Expected dict but received {type(result)}")



"""Classes related to database endpoints
"""

from __future__ import annotations

from logging import getLogger
from typing import Any, ClassVar, Optional

from metabase_tools.endpoints.generic_endpoint import Endpoint
from metabase_tools.models.database_model import DatabaseItem
from metabase_tools.models.generic_model import MissingParam
from metabase_tools.utils.logging_utils import log_call

logger = getLogger(__name__)


class Databases(Endpoint[DatabaseItem]):
    """Database related endpoint methods"""

    _BASE_EP: ClassVar[str] = "/database"
    _STD_OBJ: ClassVar[type] = DatabaseItem
    _required_params: ClassVar[list[str]] = ["name", "engine", "details"]

    @log_call
    def get(self, targets: Optional[list[int]] = None) -> list[DatabaseItem]:
        """Fetch list of databases

        Args:
            targets (list[int], optional): If provided, the list of databases to fetch

        Returns:
            list[DatabaseItem]
        """
        return super().get(targets=targets)

    def _make_create(self, **kwargs: Any) -> DatabaseItem:
        """Makes create request

        Args:
            self (DatabaseItem)

        Returns:
            DatabaseItem: self
        """
        return super()._make_create(**kwargs)

    @log_call
    def create(
        self,
        name: str | MissingParam = MissingParam(),
        engine: str | MissingParam = MissingParam(),
        details: dict[str, Any] | MissingParam = MissingParam(),
        is_full_sync: Optional[bool | MissingParam] = MissingParam(),
        is_on_demand: Optional[bool | MissingParam] = MissingParam(),
        schedules: Optional[dict[str, Any] | MissingParam] = MissingParam(),
        auto_run_queries: Optional[bool | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> DatabaseItem:
        """Create a new database

        Args:
            name (str, optional)
            engine (str, optional)
            details (dict[str, Any], optional)
            is_full_sync (bool, optional)
            is_on_demand (bool, optional)
            schedules (dict[str, Any], optional)
            auto_run_queries (bool, optional)

        Returns:
            DatabaseItem
        """
        return super()._make_create(
            name=name,
            engine=engine,
            details=details,
            is_full_sync=is_full_sync,
            is_on_demand=is_on_demand,
            schedules=schedules,
            auto_run_queries=auto_run_queries,
            **kwargs,
        )

    @log_call
    def search(
        self,
        search_params: list[dict[str, Any]],
        search_list: Optional[list[DatabaseItem]] = None,
    ) -> list[DatabaseItem]:
        """Method to search a list of databases meeting a list of parameters

        Args:
            search_params (list[dict]): Each dict contains search criteria and returns\
                 1 result
            search_list (list[DatabaseItem], optional): Provide to search an existing \
                list, by default pulls from API

        Returns:
            list[DatabaseItem]: List of databases of the relevant type
        """
        return super().search(search_params=search_params, search_list=search_list)



"""Module containing potentially useful object models"""

from .card_model import CardItem, CardQueryResult, CardRelatedObjects
from .collection_model import CollectionItem
from .database_model import DatabaseItem
from .server_settings import ServerSettings, Setting
from .user_model import UserItem

__all__ = (
    "CardItem",
    "CardQueryResult",
    "CardRelatedObjects",
    "CollectionItem",
    "DatabaseItem",
    "UserItem",
    "ServerSettings",
    "Setting",
)



"""Generic methods for endpoints
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from logging import getLogger
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Literal,
    Optional,
    TypeVar,
    cast,
)

from metabase_tools.exceptions import MetabaseApiException
from metabase_tools.models.generic_model import Item, MissingParam
from metabase_tools.utils.logging_utils import log_call

if TYPE_CHECKING:
    from metabase_tools import MetabaseApi

T = TypeVar("T", bound=Item)

logger = getLogger(__name__)


class Endpoint(ABC, Generic[T]):
    """Abstract base class for endpoints"""

    _BASE_EP: ClassVar[str]
    _STD_OBJ: ClassVar[type]
    _required_params: ClassVar[list[str]]
    _adapter: MetabaseApi

    def __init__(self, adapter: MetabaseApi):
        self._adapter = adapter

    @log_call
    def _request_list(
        self,
        http_verb: Literal["GET", "POST", "PUT", "DELETE"],
        endpoint: str,
        source: list[int],
    ) -> list[dict[str, Any]]:
        """Sends requests to API based on a list of objects

        Args:
            http_method (str): GET or POST or PUT or DELETE
            endpoint (str): Endpoint to use for request
            source (list[int] | list[dict]): List of targets or payloads

        Raises:
            InvalidParameters: Item in source is not an int or dict
            EmptyDataReceived: No data returned

        Returns:
            list[dict]: Aggregated results of all API calls
        """
        results = []

        for item in source:
            if isinstance(item, int):
                result = self._adapter.generic_request(
                    http_verb=http_verb, endpoint=endpoint.format(id=item)
                )
            else:
                raise TypeError(f"Expected list[int] but found {type(item)} in list")
            if isinstance(result, dict):
                results.append(result)
            if isinstance(result, list) and all(
                isinstance(item, dict) for item in result
            ):
                results.extend(result)
        if len(results) > 0:
            return results
        raise TypeError("Received empty list")

    @abstractmethod
    def get(self, targets: Optional[list[int]] = None) -> list[T]:
        """Generic method for returning an object or list of objects

        Args:
            targets (list[int], optional): IDs of the objects being requested

        Raises:
            InvalidParameters: Targets are not None or list[int]
            EmptyDataReceived: No data is received from the API

        Returns:
            list[T]: List of objects of the relevant type
        """
        if isinstance(targets, list) and all(isinstance(t, int) for t in targets):
            results = self._request_list(
                http_verb="GET",
                endpoint=self._BASE_EP + "/{id}",
                source=targets,
            )
            objs = [self._STD_OBJ(**result) for result in results]
            for obj in objs:
                obj.set_adapter(self._adapter)
            return objs

        if targets is None:
            # If no targets are provided, all objects of that type should be returned
            result = self._adapter.get(endpoint=self._BASE_EP)
            if isinstance(result, list):  # Validate data was returned
                # Unpack data into instances of the class and return
                objs = [
                    self._STD_OBJ(**record)
                    for record in result
                    if isinstance(record, dict)
                ]
                for obj in objs:
                    obj.set_adapter(self._adapter)
                return objs
        else:
            # If something other than None, list[int], raise error
            raise TypeError(f"Expected list[int] or None but received {type(targets)}")
        # If response.data was empty, raise error
        raise TypeError("Received empty list")

    @abstractmethod
    def _make_create(self, **kwargs: Any) -> T:
        """Generic method for creating an object

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            T: Object of the relevant type
        """
        # Validate all required parameters were provided
        try:
            missing_params = [
                param
                for param in self._required_params
                if isinstance(kwargs[param], MissingParam)
            ]
            if len(missing_params) > 0:
                raise MetabaseApiException(
                    f"Missing required parameters: {''.join(missing_params)}"
                )
        except KeyError as error:
            raise MetabaseApiException from error

        # Eliminate parameters that were not provided
        details = {k: v for k, v in kwargs.items() if not isinstance(v, MissingParam)}

        result = self._adapter.post(endpoint=self._BASE_EP, json=details)
        if isinstance(result, dict):
            obj = self._STD_OBJ(**result)
            obj.set_adapter(adapter=self._adapter)
            return cast(T, obj)
        raise TypeError(f"Expected dict but received {type(result)}")

    @abstractmethod
    def search(
        self,
        search_params: list[dict[str, Any]],
        search_list: Optional[list[T]] = None,
    ) -> list[T]:
        """Method to search a list of objects meeting a list of parameters

        Args:
            search_params (list[dict]): Each dict contains search criteria and returns\
                 1 result
            search_list (list[T], optional): Provide to search against an existing \
                list, by default pulls from API

        Returns:
            list[T]: List of objects of the relevant type
        """
        objs = search_list or self.get()
        results = []
        for param in search_params:
            for obj in objs:
                obj_dict = obj.dict()
                for key, value in param.items():
                    if key in obj_dict and value == obj_dict[key]:
                        results.append(obj)
        return results



"""Classes related to card endpoints
"""
from __future__ import annotations

from datetime import datetime
from logging import getLogger
from typing import TYPE_CHECKING, Any, ClassVar, Optional
from uuid import UUID

from packaging.version import Version
from pydantic import BaseModel, PrivateAttr
from pydantic.fields import Field

from metabase_tools.models.collection_model import CollectionItem
from metabase_tools.models.generic_model import Item, MissingParam
from metabase_tools.models.user_model import UserItem
from metabase_tools.utils.logging_utils import log_call

if TYPE_CHECKING:
    from metabase_tools.metabase import MetabaseApi

logger = getLogger(__name__)


class CardItem(Item):
    """Card object class with related methods"""

    _BASE_EP: ClassVar[str] = "/card/{id}"

    _adapter: Optional[MetabaseApi] = PrivateAttr(None)
    _server_version: Optional[Version] = PrivateAttr(None)

    description: Optional[str]
    archived: bool
    collection_position: Optional[int]
    table_id: Optional[int]
    result_metadata: Optional[list[dict[str, Any]]]
    creator: UserItem
    database_id: Optional[int]
    enable_embedding: bool
    collection_id: Optional[int]
    query_type: Optional[str]
    creator_id: int
    updated_at: datetime
    made_public_by_id: Optional[int]
    embedding_params: Optional[dict[str, Any]]
    cache_ttl: Optional[str]
    dataset_query: dict[str, Any]
    display: str
    last_edit_info: Optional[dict[str, Any]] = Field(alias="last-edit-info")
    visualization_settings: dict[str, Any]
    collection: Optional[CollectionItem]
    dataset: Optional[int]
    created_at: datetime
    public_uuid: Optional[UUID]
    can_write: Optional[bool]
    is_write: Optional[bool]
    dashboard_count: Optional[int]
    is_favorite: Optional[bool] = Field(alias="favorite")

    average_query_time: Optional[int]
    collection_preview: Optional[bool]
    entity_id: Optional[str]
    last_query_start: Optional[datetime]
    moderation_reviews: Optional[list[Any]]
    parameter_mappings: Optional[list[Any]]
    parameters: Optional[list[Any]]

    @log_call
    def set_adapter(self, adapter: MetabaseApi) -> None:
        """Sets the adapter on an object

        Args:
            adapter (MetabaseApi): Connection to MetabaseApi
        """
        super().set_adapter(adapter=adapter)

    @log_call
    def refresh(self: CardItem) -> CardItem:
        """Returns refreshed copy of the card

        Returns:
            CardItem: self
        """
        return super().refresh()

    @log_call
    def delete(self: CardItem) -> None:
        """DEPRECATED; use archive instead"""
        raise NotImplementedError

    def _make_update(self: CardItem, **kwargs: Any) -> CardItem:
        """Makes update request

        Args:
            self (CardItem)

        Returns:
            CardItem: self
        """
        return super()._make_update(**kwargs)

    @log_call
    def update(
        self: CardItem,
        visualization_settings: Optional[
            dict[str, str] | MissingParam
        ] = MissingParam(),
        description: Optional[str | MissingParam] = MissingParam(),
        archived: Optional[bool | MissingParam] = MissingParam(),
        collection_position: Optional[int | MissingParam] = MissingParam(),
        result_metadata: Optional[list[dict[str, str]] | MissingParam] = MissingParam(),
        metadata_checksum: Optional[str | MissingParam] = MissingParam(),
        enable_embedding: Optional[bool | MissingParam] = MissingParam(),
        collection_id: Optional[int | MissingParam] = MissingParam(),
        name: Optional[str | MissingParam] = MissingParam(),
        embedding_params: Optional[dict[str, str] | MissingParam] = MissingParam(),
        dataset_query: Optional[dict[str, Any] | MissingParam] = MissingParam(),
        display: Optional[str | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> CardItem:
        """Updates a card using the provided parameters

        Args:
            self (CardItem)
            visualization_settings (dict[str, str], optional)
            description (str, optional)
            archived (bool, optional)
            collection_position (int, optional)
            result_metadata (list[dict[str, str]], optional)
            metadata_checksum (str, optional)
            enable_embedding (bool, optional)
            collection_id (int, optional)
            name (str, optional)
            embedding_params (dict[str, str], optional)
            dataset_query (dict[str, Any], optional)
            display (str, optional)

        Returns:
            CardItem: _description_
        """
        return self._make_update(
            visualization_settings=visualization_settings,
            description=description,
            archived=archived,
            collection_position=collection_position,
            result_metadata=result_metadata,
            metadata_checksum=metadata_checksum,
            enable_embedding=enable_embedding,
            collection_id=collection_id,
            name=name,
            embedding_params=embedding_params,
            dataset_query=dataset_query,
            display=display,
            **kwargs,
        )

    @log_call
    def archive(self: CardItem) -> CardItem:
        """Method for archiving a card

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            CardItem: Object of the relevant type
        """
        return super().archive()

    @log_call
    def unarchive(self: CardItem) -> CardItem:
        """Method for unarchiving a card

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            CardItem: Object of the relevant type
        """
        return super().unarchive()

    @log_call
    def related(self: CardItem) -> CardRelatedObjects:
        """Objects related to target

        Returns:
            CardRelatedObjects
        """
        if self._adapter:
            result = self._adapter.get(endpoint=f"/card/{self.id}/related")
            if isinstance(result, dict):
                result["card_id"] = self.id
                return CardRelatedObjects(**result)
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def favorite(self: CardItem) -> CardItem:
        """Mark card as favorite

        Returns:
            dict: Result of favoriting operation
        """
        if self._server_version and self._server_version >= Version("v0.40"):
            raise NotImplementedError("This function was deprecated in Metabase v0.40")
        if self._adapter:
            result = self._adapter.post(endpoint=f"/card/{self.id}/favorite")
            if isinstance(result, dict):
                return self.refresh()
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def unfavorite(self: CardItem) -> CardItem:
        """Unfavorite card

        Returns:
            dict: Result of unfavoriting operation
        """
        if self._server_version and self._server_version >= Version("v0.40"):
            raise NotImplementedError("This function was deprecated in Metabase v0.40")
        if self._adapter:
            result = self._adapter.delete(endpoint=f"/card/{self.id}/favorite")
            if isinstance(result, dict):
                return self.refresh()
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def share(self: CardItem) -> CardItem:
        """Generate publicly-accessible link for card

        Returns:
            dict: UUID to be used in public link.
        """
        if self._adapter:
            result = self._adapter.post(endpoint=f"/card/{self.id}/public_link")
            if isinstance(result, dict) and "uuid" in result:
                return self.refresh()
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def unshare(self: CardItem) -> CardItem:
        """Remove publicly-accessible links for card

        Returns:
            dict: UUID to be used in public link.
        """
        if self._adapter:
            result = self._adapter.delete(endpoint=f"/card/{self.id}/public_link")
            if isinstance(result, dict) and isinstance(self.id, int):
                card = self._adapter.cards.get([self.id])[0]
                return card
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def query(self: CardItem) -> CardQueryResult:
        """Execute a query stored in card(s)

        Returns:
            CardQueryResult: Results of query
        """
        if self._adapter:
            result = self._adapter.post(endpoint=f"/card/{self.id}/query")
            if isinstance(result, dict):
                return CardQueryResult(**result)
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")


class CardQueryResult(BaseModel):
    """Object for results of a card query"""

    data: dict[str, Any]
    database_id: int
    started_at: datetime
    json_query: dict[str, Any]
    average_execution_time: Optional[int]
    status: str
    context: str
    row_count: int
    running_time: int


class CardRelatedObjects(BaseModel):
    """Objects related to the specified card"""

    card_id: int
    table: Optional[str]
    metrics: list[dict[str, int]]
    segments: list[dict[str, int]]
    dashboard_mates: list[dict[str, int]] = Field(alias="dashboard-mates")
    similar_questions: list[dict[str, int]] = Field(alias="similar-questions")
    canonical_metric: Optional[str] = Field(alias="canonical-metric")
    dashboards: list[dict[str, int]]
    collections: list[dict[str, int]]



"""Unofficial API wrapper for Metabase
"""

__version__ = "0.12.0"

from metabase_tools.exceptions import MetabaseApiException
from metabase_tools.metabase import MetabaseApi

__all__ = ("MetabaseApiException", "MetabaseApi")



"""Classes related to alert endpoints
"""
from __future__ import annotations

from datetime import datetime
from logging import getLogger
from typing import TYPE_CHECKING, Any, ClassVar, Optional

from packaging.version import Version
from pydantic import PrivateAttr
from pydantic.fields import Field

from metabase_tools.exceptions import MetabaseApiException
from metabase_tools.models.generic_model import Item, MissingParam
from metabase_tools.models.user_model import UserItem
from metabase_tools.utils.logging_utils import log_call

if TYPE_CHECKING:
    from metabase_tools.metabase import MetabaseApi

logger = getLogger(__name__)


class AlertItem(Item):
    """Alert object class with related methods"""

    _BASE_EP: ClassVar[str] = "/alert/{id}"

    _adapter: Optional[MetabaseApi] = PrivateAttr(None)
    _server_version: Optional[Version] = PrivateAttr(None)

    archived: bool
    collection_position: Optional[int]
    creator: UserItem
    can_write: Optional[bool]
    channels: list[dict[str, Any]]
    alert_condition: str
    collection_id: Optional[int]
    creator_id: int
    card: dict[str, Any]
    updated_at: datetime
    alert_first_only: bool
    lower_name: Optional[str] = Field(alias="lower-name")
    entity_id: Optional[str]
    skip_if_empty: bool
    parameters: list[dict[str, Any]]
    dashboard_id: Optional[int]
    created_at: datetime
    alert_above_goal: Optional[bool]

    @log_call
    def set_adapter(self, adapter: MetabaseApi) -> None:
        """Sets the adapter on an object

        Args:
            adapter (MetabaseApi): Connection to MetabaseApi
        """
        super().set_adapter(adapter=adapter)

    @log_call
    def refresh(self: AlertItem) -> AlertItem:
        """Returns refreshed copy of the alert

        Returns:
            AlertItem: self
        """
        if self._adapter and self._adapter.server_version >= Version("v0.41"):
            return super().refresh()
        if self._adapter and isinstance(self.id, int):
            return self._adapter.alerts.get(targets=[self.id])[0]
        raise MetabaseApiException

    @log_call
    def delete(self: AlertItem) -> None:
        """DEPRECATED; use archive instead"""
        raise NotImplementedError

    def _make_update(self: AlertItem, **kwargs: Any) -> AlertItem:
        """Makes update request

        Args:
            self (AlertItem)

        Returns:
            AlertItem: self
        """
        return super()._make_update(**kwargs)

    @log_call
    def update(
        self: AlertItem,
        alert_condition: Optional[str | MissingParam] = MissingParam(),
        alert_first_only: Optional[bool | MissingParam] = MissingParam(),
        alert_above_goal: Optional[bool | MissingParam] = MissingParam(),
        card: Optional[dict[str, Any] | MissingParam] = MissingParam(),
        channels: Optional[list[dict[str, Any]] | MissingParam] = MissingParam(),
        archived: Optional[bool | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> AlertItem:
        """Updates a card using the provided parameters

        Args:
            self (AlertItem): _description_
            alert_condition (str, optional)
            alert_first_only (bool, optional)
            alert_above_goal (bool, optional)
            card (dict[str, Any], optional)
            channels (list[dict[str, Any]], optional)
            archived (bool, optional)

        Returns:
            AlertItem: _description_
        """
        return self._make_update(
            alert_condition=alert_condition,
            alert_first_only=alert_first_only,
            alert_above_goal=alert_above_goal,
            card=card,
            channels=channels,
            archived=archived,
            **kwargs,
        )

    @log_call
    def archive(self: AlertItem) -> AlertItem:
        """Method for archiving an alert

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            AlertItem: Object of the relevant type
        """
        return super().archive()

    @log_call
    def unarchive(self: AlertItem) -> AlertItem:
        """Method for unarchiving an alert

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            AlertItem: Object of the relevant type
        """
        return super().unarchive()



"""
MetabaseTools extends MetabaseApi with additional complex functions
"""
from __future__ import annotations  # Included for support of |

from json import dumps, loads
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from metabase_tools.exceptions import MetabaseApiException
from metabase_tools.models.card_model import CardItem
from metabase_tools.utils.logging_utils import log_call

if TYPE_CHECKING:
    from metabase_tools.metabase import MetabaseApi

logger = getLogger(__name__)


class MetabaseTools:
    """Extends MetabaseApi with additional complex functions"""

    _adapter: MetabaseApi

    def __init__(self, adapter: MetabaseApi):
        self._adapter = adapter

    @log_call
    def download_native_queries(
        self,
        save_file: Optional[Path | str] = None,
        root_folder: Path | str = ".",
        file_extension: str = "sql",
    ) -> Path:
        """Downloads all native queries into a JSON file

        Args:
            save_file (Path | str, optional): Path to save mapping file, defaults to \
                mapping_{timestamp}.json
            root_folder (Path | str, optional): Root folder to save queries, by \
                default "."
            file_extension (str, optional): File extension to save the queries, by \
                default "sql"

        Returns:
            Path: Path to save file
        """
        # Determine save path
        root_folder = Path(root_folder)  # Convert root folder to a path object
        save_file = Path(save_file or "mapping.json")

        # Download list of cards from Metabase API and filter to only native queries
        cards = [
            card
            for card in self._adapter.cards.get()
            if (
                card.query_type == "native"
                and card.collection
                and card.collection.personal_owner_id is None
            )
        ]
        logger.debug("Found %s cards with native queries", len(cards))

        # Create dictionary of collections for file paths
        collections_by_id = self._get_collections_dict(key="id")
        logger.debug("Generated flat list of %s collections", len(collections_by_id))

        # Format filtered list
        cards_to_write = []

        for card in cards:
            new_card = self._get_mapping_details(
                card, collections_by_id=collections_by_id
            )
            cards_to_write.append(new_card)

            try:
                self._save_query(
                    card=card,
                    save_path=f"{root_folder}/{new_card['path']}",
                    file_extension=file_extension,
                )
            except OSError:
                logger.warning("Skipping %s (name error)", card.name)
                continue
            logger.debug(
                "%s saved to %s",
                card.name,
                f"{root_folder}/{new_card['path']}",
            )

        # Save mapping file
        mapping_path = Path(f"{root_folder}")
        mapping_path.mkdir(parents=True, exist_ok=True)
        mapping_path /= save_file
        logger.debug("Completed iterating through list, saving file: %s", mapping_path)
        with open(mapping_path, "w", newline="", encoding="utf-8") as file:
            file.write(dumps(cards_to_write, indent=2))

        # Returns path to file saved
        return mapping_path

    @log_call
    def upload_native_queries(
        self,
        mapping_path: Path | str,
        file_extension: str,
        dry_run: bool = True,
        stop_on_error: bool = False,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """Uploads queries to Metabase

        Args:
            mapping_path (Path | str): Path to the mapping configuration file, by \
                default None
            dry_run (bool, optional): Execute task as a dry run (i.e. do not make \
                any changes), by default True
            stop_on_error (bool, optional): Raise error and stop if an error is \
                encountered. Defaults to False.

        Raises:
            FileNotFoundError: The file referenced was not found

        Returns:
            list[dict] | dict: Results of upload
        """
        # Open mapping configuration file
        mapping_path = Path(mapping_path or "./mapping.json")
        with open(mapping_path, "r", newline="", encoding="utf-8") as file:
            cards = loads(file.read())

        # Iterate through mapping file
        changes: dict[str, list[dict[str, Any]]] = {
            "updates": [],
            "creates": [],
            "errors": [],
        }
        collections_by_path = self._get_collections_dict(key="path")
        for card in cards:
            card_path = Path(
                f"{mapping_path.parent}/{card['path']}/{card['name']}.{file_extension}"
            )
            if card_path.exists():
                # Check if a card with the same name exists in the listed location
                dev_coll_id = collections_by_path[card["path"]]["id"]
                try:
                    collection = self._adapter.collections.get(targets=[dev_coll_id])[0]
                    card_id = collection.check_for_object(card["name"])
                except MetabaseApiException:  # No items in collection or not found
                    logger.debug(
                        "%s not found in listed location, creating", card["name"]
                    )
                    card_id = None

                if card_id:  # update existing card
                    card_result = self._update_existing_card(
                        card_id=card_id, card_path=card_path
                    )
                    if card_result:
                        changes["updates"].append(card_result)
                else:  # create card
                    card_result = self._create_new_card(
                        card=card, card_path=card_path, dev_coll_id=dev_coll_id
                    )
                    if card_result:
                        changes["creates"].append(card_result)

            else:  # File does not exist
                if stop_on_error:
                    logger.error("Unable to process %s (file not found)", card["name"])
                    raise FileNotFoundError(f"{card_path} not found")
                logger.warning("Skipping %s (file not found)", card["name"])
                changes["errors"].append(card)

        # Loop exit before pushing changes to Metabase in case errors are encountered
        # Push changes back to Metabase API
        if not dry_run:
            return self._execute_changes(changes)
        return changes

    @log_call
    def _execute_changes(
        self, changes: dict[str, list[dict[str, Any]]]
    ) -> list[dict[str, Any]] | dict[str, Any]:
        results = []
        if len(changes["updates"]) > 0:
            update_results = []
            for update in changes["updates"]:
                card = self._adapter.cards.get([update["id"]])[0]
                update_results.append(card.update(**update))

            if isinstance(update_results, list):
                for result in update_results:
                    results.append(
                        {"id": result.id, "name": result.name, "is_success": True}
                    )

        create_results = []
        if len(changes["creates"]) > 0:
            create_results = [
                self._adapter.cards.create(**details) for details in changes["creates"]
            ]
            if isinstance(create_results, list):
                for result in create_results:
                    results.append(
                        {"id": result.id, "name": result.name, "is_success": True}
                    )

        return results

    @log_call
    def _get_mapping_details(
        self, card: CardItem, collections_by_id: dict[Any, Any]
    ) -> dict[str, Any]:
        mapping_details = {
            "name": card.name,
            "collection_id": card.collection_id,
            "path": collections_by_id[card.collection_id]["path"],
        }

        mapping_details["database_id"] = card.database_id
        mapping_details["database_name"] = self._adapter.databases.search(
            search_params=[{"id": card.database_id}]
        )[0].name

        return mapping_details

    @log_call
    def _save_query(self, card: CardItem, save_path: str, file_extension: str) -> None:
        # SQL file creation
        sql_code = card.dataset_query["native"]["query"]
        sql_path = Path(f"{save_path}")
        sql_path.mkdir(parents=True, exist_ok=True)
        sql_path /= f"{card.name}.{file_extension}"
        with open(sql_path, "w", newline="", encoding="utf-8") as file:
            file.write(sql_code)

    @log_call
    def _get_collections_dict(self, key: str) -> dict[Any, Any]:
        collections = self._adapter.collections.get_flat_list()
        non_keys = [k for k in collections[0].keys() if k != key]
        return {item[key]: {nk: item[nk] for nk in non_keys} for item in collections}

    @log_call
    def _update_existing_card(
        self,
        card_id: int,
        card_path: Path | str,
    ) -> dict[str, Any]:
        prod_card = self._adapter.cards.get(targets=[card_id])[0]
        with open(card_path, "r", newline="", encoding="utf-8") as file:
            dev_code = file.read()
        if dev_code != prod_card.dataset_query["native"]["query"]:
            dev_query = prod_card.dataset_query.copy()
            dev_query["native"]["query"] = dev_code
            dev_def = {"id": card_id, "dataset_query": dev_query}
            return dev_def
        return {}

    @log_call
    def _create_new_card(
        self, card: dict[str, Any], card_path: Path, dev_coll_id: int
    ) -> dict[str, Any]:
        with open(card_path, "r", newline="", encoding="utf-8") as file:
            dev_query = file.read()
        db_id = [
            database.id
            for database in self._adapter.databases.get()
            if card["database_name"] == database.name
        ][0]

        new_card_def = {
            "visualization_settings": {},
            "collection_id": dev_coll_id,
            "name": card["name"],
            "dataset_query": {
                "type": "native",
                "native": {"query": dev_query},
                "database": db_id,
            },
            "display": "table",
        }
        return new_card_def



"""Classes related to database endpoints
"""
from __future__ import annotations

from datetime import datetime
from logging import getLogger
from typing import TYPE_CHECKING, Any, ClassVar, Optional

from pydantic import Field, PrivateAttr

from metabase_tools.models.generic_model import Item, MissingParam
from metabase_tools.utils.logging_utils import log_call

if TYPE_CHECKING:
    from metabase_tools.metabase import MetabaseApi

logger = getLogger(__name__)


class DatabaseItem(Item):
    """Database object class with related methods"""

    _BASE_EP: ClassVar[str] = "/database/{id}"

    _adapter: Optional[MetabaseApi] = PrivateAttr(None)

    description: Optional[str]
    features: list[str]
    cache_field_values_schedule: str
    timezone: Optional[str]
    auto_run_queries: bool
    metadata_sync_schedule: str
    caveats: Optional[str]
    is_full_sync: bool
    updated_at: datetime
    native_permissions: Optional[str]
    details: dict[str, Any]
    is_sample: bool
    is_on_demand: bool
    options: Optional[str]
    engine: str
    refingerprint: Optional[str]
    created_at: datetime
    points_of_interest: Optional[str]
    schedules: Optional[dict[str, Any]]
    cache_ttl: Optional[int]
    creator_id: Optional[int]
    initial_sync_status: Optional[str]
    settings: Optional[Any]
    can_manage: Optional[bool] = Field(alias="can-manage")

    @log_call
    def refresh(self: DatabaseItem) -> DatabaseItem:
        """Returns refreshed copy of the database

        Returns:
            DatabaseItem: self
        """
        return super().refresh()

    @log_call
    def delete(self: DatabaseItem) -> None:
        """Deletes the database"""
        return super().delete()

    def _make_update(self: DatabaseItem, **kwargs: Any) -> DatabaseItem:
        """Makes update request

        Args:
            self (DatabaseItem)

        Returns:
            DatabaseItem: self
        """
        return super()._make_update(**kwargs)

    @log_call
    def update(
        self: DatabaseItem,
        engine: Optional[str | MissingParam] = MissingParam(),
        schedules: Optional[dict[str, Any] | MissingParam] = MissingParam(),
        refingerprint: Optional[bool | MissingParam] = MissingParam(),
        points_of_interest: Optional[str | MissingParam] = MissingParam(),
        description: Optional[str | MissingParam] = MissingParam(),
        name: Optional[str | MissingParam] = MissingParam(),
        caveats: Optional[str | MissingParam] = MissingParam(),
        cache_ttl: Optional[int | MissingParam] = MissingParam(),
        details: Optional[dict[str, Any] | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> DatabaseItem:
        """Updates a database using the provided parameters

        Args:
            self (DatabaseItem)
            engine (str, optional)
            schedules (dict[str, Any], optional)
            refingerprint (bool, optional)
            points_of_interest (str, optional)
            description (str, optional)
            name (str, optional)
            caveats (str, optional)
            cache_ttl (int, optional)
            details (dict[str, Any], optional)

        Returns:
            DatabaseItem
        """
        return self._make_update(
            engine=engine,
            schedules=schedules,
            refingerprint=refingerprint,
            points_of_interest=points_of_interest,
            description=description,
            name=name,
            caveats=caveats,
            cache_ttl=cache_ttl,
            details=details,
            **kwargs,
        )



"""Classes related to collections endpoints
"""
from __future__ import annotations  # Included for support of |

from datetime import datetime
from logging import getLogger
from typing import TYPE_CHECKING, Any, ClassVar, Optional

from pydantic import PrivateAttr

from metabase_tools.exceptions import MetabaseApiException
from metabase_tools.models.generic_model import Item, MissingParam
from metabase_tools.utils.logging_utils import log_call

if TYPE_CHECKING:
    from metabase_tools.metabase import MetabaseApi

logger = getLogger(__name__)


class CollectionItem(Item):
    """Collection object class with related methods"""

    _BASE_EP: ClassVar[str] = "/collection/{id}"

    _adapter: Optional[MetabaseApi] = PrivateAttr(None)

    description: Optional[str]
    archived: Optional[bool]
    slug: Optional[str]
    color: Optional[str]
    personal_owner_id: Optional[int]
    location: Optional[str]
    namespace: Optional[int]
    effective_location: Optional[str]
    effective_ancestors: Optional[list[dict[str, Any]]]
    can_write: Optional[bool]
    parent_id: Optional[int]

    authority_level: Optional[Any]
    entity_id: Optional[str]
    created_at: Optional[datetime]

    @log_call
    def set_adapter(self, adapter: MetabaseApi) -> None:
        """Sets the adapter on an object

        Args:
            adapter (MetabaseApi): Connection to MetabaseApi
        """
        super().set_adapter(adapter=adapter)

    @log_call
    def refresh(self: CollectionItem) -> CollectionItem:
        """Returns refreshed copy of the collection

        Returns:
            CollectionItem: self
        """
        return super().refresh()

    def _make_update(self: CollectionItem, **kwargs: Any) -> CollectionItem:
        """Makes update request

        Args:
            self (CollectionItem)

        Returns:
            CollectionItem: self
        """
        return super()._make_update(**kwargs)

    @log_call
    def update(
        self: CollectionItem,
        name: Optional[str | MissingParam] = MissingParam(),
        color: Optional[str | MissingParam] = MissingParam(),
        description: Optional[str | MissingParam] = MissingParam(),
        archived: Optional[bool | MissingParam] = MissingParam(),
        parent_id: Optional[int | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> CollectionItem:
        """Updates a collection using the provided parameters

        Args:
            self (CollectionItem)
            name (str, optional)
            color (str, optional)
            description (str, optional)
            archived (bool, optional)
            parent_id (int, optional)

        Returns:
            CollectionItem
        """
        return self._make_update(
            name=name,
            color=color,
            description=description,
            archived=archived,
            parent_id=parent_id,
            **kwargs,
        )

    @log_call
    def archive(self: CollectionItem) -> CollectionItem:
        """Method for archiving a collection

        Returns:
            CollectionItem: Object of the relevant type
        """
        return super().archive()

    @log_call
    def unarchive(self: CollectionItem) -> CollectionItem:
        """Method for unarchiving a collection

        Returns:
            CollectionItem: Object of the relevant type
        """
        return super().unarchive()

    @log_call
    def delete(self: CollectionItem) -> None:
        """DEPRECATED; use archive instead"""
        raise NotImplementedError

    @log_call
    def get_contents(
        self,
        model_type: Optional[str] = None,
        archived: bool = False,
    ) -> list[dict[str, Any]]:
        """Get the contents of the provided collection

        Args:
            adapter (MetabaseApi): Connection to Metabase API
            collection_id (int): ID of the requested collection
            model_type (str, optional): Filter to provided model. Defaults to all.
            archived (bool, optional): Archived objects. Defaults to False.

        Raises:
            EmptyDataReceived: No results from API

        Returns:
            list: Contents of collection
        """
        params: dict[Any, Any] = {}
        if archived:
            params["archived"] = archived
        if model_type:
            params["model"] = model_type

        if self._adapter:
            result = self._adapter.get(
                endpoint=f"/collection/{self.id}/items",
                params=params,
            )
            if isinstance(result, list) and all(
                isinstance(record, dict) for record in result
            ):
                return result
            raise TypeError(f"Expected list[dict], received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def check_for_object(self, item_name: str) -> int:
        """Checks for object in the collection and returns the id, if found

        Args:
            item_name (str): Name of the item being located

        Raises:
            MetabaseApiException: Item not found

        Returns:
            int: id of the item being located
        """
        collection_items = self.get_contents(model_type="card", archived=False)
        for item in collection_items:
            if item["name"] == item_name and isinstance(item["id"], int):
                return item["id"]
        raise MetabaseApiException



"""Classes related to user endpoints
"""
from __future__ import annotations

from datetime import datetime
from logging import getLogger
from typing import TYPE_CHECKING, Any, ClassVar, Optional

from packaging.version import Version
from pydantic.fields import Field, PrivateAttr

from metabase_tools.models.generic_model import Item, MissingParam
from metabase_tools.utils.logging_utils import log_call

if TYPE_CHECKING:
    from metabase_tools.metabase import MetabaseApi

logger = getLogger(__name__)


class UserItem(Item):
    """User object class with related methods"""

    _BASE_EP: ClassVar[str] = "/user/{id}"

    _adapter: Optional[MetabaseApi] = PrivateAttr(None)

    name: str = Field(alias="common_name")
    email: str
    first_name: str
    last_name: str
    date_joined: datetime
    last_login: Optional[datetime]
    updated_at: Optional[datetime]
    is_qbnewb: bool
    is_superuser: bool
    ldap_auth: Optional[bool]
    google_auth: Optional[bool]
    is_active: Optional[bool]
    locale: Optional[str]
    group_ids: Optional[list[int]]
    login_attributes: Optional[list[dict[str, Any]]]
    personal_collection_id: Optional[int]
    has_invited_second_user: Optional[bool]
    user_group_memberships: Optional[list[dict[str, int]]]
    first_login: Optional[datetime]
    has_question_and_dashboard: Optional[bool]
    is_installer: Optional[bool]
    sso_source: Optional[str]

    @log_call
    def refresh(self: UserItem) -> UserItem:
        """Returns refreshed copy of the user

        Returns:
            UserItem: self
        """
        return super().refresh()

    @log_call
    def disable(self) -> None:
        """Disables user"""
        return super().delete()

    @log_call
    def enable(self) -> UserItem:
        """Enable user

        Returns:
            UserItem: Enabled users
        """
        if self._adapter:
            result = self._adapter.put(endpoint=f"/user/{self.id}/reactivate")
            if isinstance(result, dict):
                obj = self.__class__(**result)
                obj.set_adapter(adapter=self._adapter)
                return obj
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def resend_invite(self) -> dict[str, bool]:
        """Resent user invites

        Returns:
            UserItem: User with a resent invite
        """
        if self._adapter:
            result = self._adapter.post(endpoint=f"/user/{self.id}/send_invite")
            if isinstance(result, dict):
                return result
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def update_password(self: UserItem, payload: dict[str, Any]) -> UserItem:
        """Updates passwords for users

        Args:
            payload (dict): New password

        Returns:
            UserItem: User with reset passwords
        """
        if self._adapter:
            result = self._adapter.put(
                endpoint=f"/user/{self.id}/password", json=payload
            )
            if isinstance(result, dict):
                obj = self.__class__(**result)
                obj.set_adapter(adapter=self._adapter)
                return obj
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def qbnewb(self) -> dict[str, bool]:
        """Indicate that a user has been informed about Query Builder.

        Returns:
            UserItem: User with query builder toggle set
        """
        if self._server_version and self._server_version >= Version("v0.42"):
            raise NotImplementedError("This function was deprecated in Metabase v0.42")
        if self._adapter:
            result = self._adapter.put(endpoint=f"/user/{self.id}/qbnewb")
            if isinstance(result, dict):
                return result
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    def _make_update(self: UserItem, **kwargs: Any) -> UserItem:
        """Makes update request

        Args:
            self (UserItem)

        Returns:
            UserItem: self
        """
        return super()._make_update(**kwargs)

    @log_call
    def update(
        self: UserItem,
        email: Optional[str | MissingParam] = MissingParam(),
        first_name: Optional[str | MissingParam] = MissingParam(),
        is_group_manager: Optional[bool | MissingParam] = MissingParam(),
        locale: Optional[str | MissingParam] = MissingParam(),
        user_group_memberships: Optional[list[int] | MissingParam] = MissingParam(),
        is_superuser: Optional[bool | MissingParam] = MissingParam(),
        login_attributes: Optional[str | MissingParam] = MissingParam(),
        last_name: Optional[str | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> UserItem:
        """Updates a user using the provided parameters

        Args:
            self (UserItem)
            email (str, optional)
            first_name (str, optional)
            is_group_manager (bool, optional)
            locale (str, optional)
            user_group_memberships (list[int], optional)
            is_superuser (bool, optional)
            login_attributes (str, optional)
            last_name (str, optional)

        Returns:
            UserItem
        """
        return self._make_update(
            email=email,
            first_name=first_name,
            is_group_manager=is_group_manager,
            locale=locale,
            user_group_memberships=user_group_memberships,
            is_superuser=is_superuser,
            login_attributes=login_attributes,
            last_name=last_name,
            **kwargs,
        )



"""
Rest adapter for the Metabase API
"""
from __future__ import annotations

from json import JSONDecodeError
from logging import getLogger
from pathlib import Path
from typing import Any, Literal, Optional

from packaging.version import Version
from requests import Response, Session
from requests.exceptions import RequestException

from metabase_tools.endpoints.alerts_endpoint import Alerts
from metabase_tools.endpoints.cards_endpoint import Cards
from metabase_tools.endpoints.collections_endpoint import Collections
from metabase_tools.endpoints.databases_endpoint import Databases
from metabase_tools.endpoints.users_endpoint import Users
from metabase_tools.exceptions import MetabaseApiException
from metabase_tools.models.server_settings import ServerSettings, Setting
from metabase_tools.tools.tools import MetabaseTools

logger = getLogger(__name__)


class MetabaseApi:
    """Metabase API adapter"""

    server_version: Version

    cards: Cards
    collections: Collections
    databases: Databases
    settings: ServerSettings
    tools: MetabaseTools
    users: Users

    def __init__(
        self,
        metabase_url: str,
        credentials: Optional[dict[str, str]] = None,
        cache_token: bool = False,
        token_path: Optional[Path | str] = None,
        session: Optional[Session] = None,
    ):
        if not credentials and not token_path:
            raise MetabaseApiException("No authentication method provided")
        credentials = credentials or {}
        token_path = Path(token_path) if token_path else None

        # Validate Metabase URL
        self.metabase_url = self._validate_base_url(url=metabase_url)

        # Starts session to be reused by the adapter so that the auth token is cached
        self._session = session or Session()

        # Authenticate
        self._authenticate(token_path=token_path, credentials=credentials)

        # Cache token, if set during init
        if cache_token:
            save_path = Path(token_path or "metabase.token")
            self._save_token(save_path=save_path)

        # Set server version for compatability checks
        self._set_server_version()

        # Create endpoints
        self.alerts = Alerts(self)
        self.cards = Cards(self)
        self.collections = Collections(self)
        self.databases = Databases(self)
        self.settings = self._fetch_settings()
        self.tools = MetabaseTools(self)
        self.users = Users(self)

    def _validate_base_url(self, url: str) -> str:
        if url[-1] == "/":
            url = url[:-1]
        if url[-4:] == "/api":
            url = url[:-4]
        if url[:4] != "http":
            url = f"http://{url}"
        return f"{url}/api"

    def _authenticate(
        self, token_path: Optional[Path], credentials: dict[str, str]
    ) -> None:
        authed = False
        # Try cached token first
        if token_path and token_path.exists():
            authed = self._auth_with_cached_token(token_path=token_path)
            if not authed:
                self._delete_cached_token(token_path=token_path)
        # Try token passed as credentials next
        if not authed and "token" in credentials:
            authed = self._auth_with_passed_token(credentials=credentials)
        # Finally try username and password
        if not authed and "username" in credentials and "password" in credentials:
            authed = self._auth_with_login(credentials=credentials)
        # Raise error if still not authenticated
        if not authed:
            logger.error("Failed to authenticate")
            raise MetabaseApiException(
                "Failed to authenticate with credentials provided"
            )

    def _add_token_to_header(self, token: str) -> None:
        headers = {
            "Content-Type": "application/json",
            "X-Metabase-Session": token,
        }
        self._session.headers.update(headers)

    def _delete_cached_token(self, token_path: Path) -> None:
        if token_path.exists():
            logger.warning("Deleting token file")
            token_path.unlink()

    def _auth_with_cached_token(self, token_path: Path) -> bool:
        with open(token_path, "r", encoding="utf-8") as file:
            token = file.read()
        logger.info("Attempting authentication with token file")
        self._add_token_to_header(token=token)
        authed = self.test_for_auth()
        logger.info(
            "Authenticated with token file"
            if authed
            else "Failed to authenticate with token file"
        )
        return authed

    def _auth_with_passed_token(self, credentials: dict[str, str]) -> bool:
        logger.info("Attempting authentication with token passed")
        self._add_token_to_header(token=credentials["token"])
        authed = self.test_for_auth()
        logger.info(
            "Authenticated with token passed"
            if authed
            else "Failed to authenticate with token passed"
        )
        return authed

    def _auth_with_login(self, credentials: dict[str, str]) -> bool:
        """Private method for authenticating a session with the API

        Args:
            credentials (dict): Username and password
        """
        try:
            logger.info("Attempting authentication with username and password")
            response = self._session.post(
                f"{self.metabase_url}/session", json=credentials
            )
            self._add_token_to_header(token=response.json()["id"])
            authed = self.test_for_auth()
            logger.info(
                "Authenticated with login"
                if authed
                else "Failed to authenticate with login"
            )
            return authed
        except KeyError as error_raised:
            logger.warning(
                "Exception encountered during attempt to authenticate with login \
                    passed: %s",
                error_raised,
            )
        return False

    def test_for_auth(self) -> bool:
        """Validates successful authentication by attempting to retrieve data about \
            the current user

        Returns:
            bool: Authentication success status
        """
        return (
            200
            <= self._session.get(f"{self.metabase_url}/user/current").status_code
            <= 299
        )

    def _save_token(self, save_path: Path | str) -> None:
        """Writes active token to the specified file

        Args:
            save_path (Path | str): Name of file to write
        """
        logger.debug("Saving token to %s", save_path)
        token = str(self._session.headers.get("X-Metabase-Session"))
        with open(save_path, "w", encoding="utf-8") as file:
            file.write(token)

    def _make_request(
        self,
        method: str,
        url: str,
        params: Optional[dict[str, Any]] = None,
        json: Optional[dict[str, Any]] = None,
    ) -> Response:
        """Perform an HTTP request, catching and re-raising any exceptions

        Args:
            method (str): GET or POST or DELETE or PUT
            url (str): URL endpoint
            params (dict, optional): Endpoint parameters
            json (dict, optional): Data payload

        Raises:
            RequestFailure: Request failed

        Returns:
            Response: Response from the API
        """
        try:
            logger.info("Making HTTP request: %s:%s:%s", method, url, params)
            return self._session.request(
                method=method, url=url, params=params, json=json
            )
        except RequestException as error_raised:
            logger.error(str(error_raised))
            raise MetabaseApiException from error_raised

    def generic_request(
        self,
        http_verb: Literal["GET", "POST", "PUT", "DELETE"],
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        json: Optional[dict[str, Any]] = None,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """Method for dispatching HTTP requests

        Args:
            http_method (str): GET or POST or PUT or DELETE
            endpoint (str): URL endpoint
            params (dict, optional): Endpoint parameters
            json (dict, optional): Data payload

        Raises:
            InvalidDataReceived: Unable to decode response from API
            AuthenticationFailure: Auth failure received from API
            RequestFailure: Other failure during request

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        log_line_post = "Request result: success=%s, status_code=%s, message=%s"
        response = self._make_request(
            method=http_verb,
            url=self.metabase_url + endpoint,
            params=params,
            json=json,
        )

        # If status_code in 200-299 range, return Result, else raise exception
        if 299 >= response.status_code >= 200:
            logger.info(log_line_post, True, response.status_code, response.reason)
            try:
                data = response.json()
                if isinstance(data, dict) and all(
                    key in data for key in ["data", "total"]
                ):
                    data = data["data"]
                if isinstance(data, (list, dict)):
                    return data
            except JSONDecodeError:
                if response.status_code == 204:
                    return {"success": True}
                logger.error(log_line_post, False, response.status_code, response.text)
                raise
        elif response.status_code == 401:
            logger.error(log_line_post, False, response.status_code, response.text)
            raise MetabaseApiException(
                f"Failed to authenticate: {response.status_code} - {response.reason}"
            )

        error_line = f"{response.status_code} - {response.reason} - {response.text}"
        logger.error(log_line_post, False, response.status_code, response.text)
        raise MetabaseApiException(error_line)

    def get(
        self, endpoint: str, params: Optional[dict[str, Any]] = None
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """HTTP GET request

        Args:
            endpoint (str): URL endpoint
            ep_params (dict, optional): Endpoint parameters

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        return self.generic_request(http_verb="GET", endpoint=endpoint, params=params)

    def post(
        self,
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        json: Optional[dict[str, Any]] = None,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """HTTP POST request

        Args:
            endpoint (str): URL endpoint
            params (dict, optional): Endpoint parameters
            json (dict, optional): Data payload

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        return self.generic_request(
            http_verb="POST", endpoint=endpoint, params=params, json=json
        )

    def delete(
        self, endpoint: str, params: Optional[dict[str, Any]] = None
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """HTTP DELETE request

        Args:
            endpoint (str): URL endpoint
            params (dict, optional): Endpoint parameters

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        return self.generic_request(
            http_verb="DELETE", endpoint=endpoint, params=params
        )

    def put(
        self,
        endpoint: str,
        params: Optional[dict[str, Any]] = None,
        json: Optional[dict[str, Any]] = None,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """HTTP PUT request

        Args:
            endpoint (str): URL endpoint
            ep_params (dict, optional): Endpoint parameters
            json (dict, optional): Data payload

        Returns:
            list[dict[str, Any]] | dict[str, Any]: Response from API
        """
        return self.generic_request(
            http_verb="PUT", endpoint=endpoint, params=params, json=json
        )

    def _set_server_version(self) -> None:
        """Get the Metabase version running on the server

        Returns:
            str: version string
        """
        properties = self.get("/session/properties")
        if isinstance(properties, dict):
            self.server_version = Version(properties["version"]["tag"])
            logger.info("Server version: %s", self.server_version)
        else:
            logger.error("Unable to fetch server version")
            raise MetabaseApiException("Unable to fetch server version")

    def _fetch_settings(self) -> ServerSettings:
        """Retrieves settings from Metabase server"""
        settings = self.get("/setting")
        if isinstance(settings, list):
            settings = {setting["key"]: Setting(**setting) for setting in settings}
            server_settings = ServerSettings(**settings)
            server_settings.set_adapter(self)
            return server_settings
        logger.error("Unable to fetch settings")
        raise MetabaseApiException("Unable to fetch settings")



"""Classes related to alert endpoints
"""

from __future__ import annotations

from logging import getLogger
from typing import Any, ClassVar, Optional

from packaging.version import Version

from metabase_tools.endpoints.generic_endpoint import Endpoint
from metabase_tools.models.alert_model import AlertItem
from metabase_tools.models.generic_model import MissingParam
from metabase_tools.utils.logging_utils import log_call

logger = getLogger(__name__)


class Alerts(Endpoint[AlertItem]):
    """Alert related endpoint methods"""

    _BASE_EP: ClassVar[str] = "/alert"
    _STD_OBJ: ClassVar[type] = AlertItem

    _required_params: ClassVar[list[str]] = [
        "alert_condition",
        "card",
        "channels",
        "alert_first_only",
        "alert_above_goal",
    ]

    @log_call
    def get(self, targets: Optional[list[int]] = None) -> list[AlertItem]:
        """Fetch list of alerts

        Args:
            targets (list[int], optional): If provided, the list of alerts to fetch

        Returns:
            list[AlertItem]
        """
        if targets and not isinstance(targets, list):
            raise TypeError
        if self._adapter and self._adapter.server_version >= Version("v0.41"):
            return super().get(targets=targets)  # standard if supported by version
        if targets and isinstance(targets, list):
            return [x for x in super().get() if x.id in targets]  # get all and filter
        return super().get()  # get all

    def get_by_card(self, targets: list[int]) -> list[AlertItem]:
        """Get all alerts for the given card IDs

        Args:
            targets (list[int]): List of card IDs

        Returns:
            list[AlertItem]: List of associated alerts
        """
        result = self._request_list(
            "GET", endpoint="/alert/question/{id}", source=targets
        )
        alerts = [AlertItem(**x) for x in result]
        for alert in alerts:
            alert.set_adapter(self._adapter)
        return alerts

    def _make_create(self, **kwargs: Any) -> AlertItem:
        """Makes create request

        Args:
            self (AlertItem)

        Returns:
            AlertItem: self
        """
        return super()._make_create(**kwargs)

    @log_call
    def create(
        self,
        visualization_settings: dict[str, Any] | MissingParam = MissingParam(),
        name: str | MissingParam = MissingParam(),
        dataset_query: dict[str, Any] | MissingParam = MissingParam(),
        display: str | MissingParam = MissingParam(),
        description: Optional[str | MissingParam] = MissingParam(),
        collection_position: Optional[int | MissingParam] = MissingParam(),
        result_metadata: Optional[list[dict[str, Any]] | MissingParam] = MissingParam(),
        metadata_checksum: Optional[str | MissingParam] = MissingParam(),
        collection_id: Optional[int | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> AlertItem:
        """Creates a new alert

        Args:
            visualization_settings (dict[str, Any])
            name (str)
            dataset_query (dict[str, Any])
            display (str)
            description (str, optional)
            collection_position (int, optional)
            result_metadata (list[dict[str, Any]], optional)
            metadata_checksum (str, optional)
            collection_id (int, optional)

        Returns:
            AlertItem
        """
        return self._make_create(
            visualization_settings=visualization_settings,
            description=description,
            collection_position=collection_position,
            result_metadata=result_metadata,
            metadata_checksum=metadata_checksum,
            collection_id=collection_id,
            name=name,
            dataset_query=dataset_query,
            display=display,
            **kwargs,
        )

    @log_call
    def search(
        self,
        search_params: list[dict[str, Any]],
        search_list: Optional[list[AlertItem]] = None,
    ) -> list[AlertItem]:
        """Method to search a list of alerts meeting a list of parameters

        Args:
            search_params (list[dict]): Each dict contains search criteria and returns\
                 1 result
            search_list (list[AlertItem], optional): Provide to search an existing \
                list, by default pulls from API

        Returns:
            list[AlertItem]: List of alerts of the relevant type
        """
        return super().search(search_params=search_params, search_list=search_list)



"""
    Generic classes for Metabase
"""
from __future__ import annotations

from abc import ABC
from logging import getLogger
from typing import TYPE_CHECKING, Any, ClassVar, Optional, TypeVar

from packaging.version import Version
from pydantic import BaseModel, PrivateAttr

from metabase_tools.utils.logging_utils import log_call

if TYPE_CHECKING:
    from metabase_tools.metabase import MetabaseApi

T = TypeVar("T", bound="Item")
logger = getLogger(__name__)


class MissingParam(BaseModel):
    """Used for parameters that were not supplied during an update or object creation"""


class Item(BaseModel, ABC, extra="forbid"):
    """Base class for all Metabase objects. Provides generic fields and methods."""

    _BASE_EP: ClassVar[str]

    _adapter: Optional[MetabaseApi] = PrivateAttr(None)
    _server_version: Optional[Version] = PrivateAttr(None)

    id: int | str
    name: Optional[str]

    @log_call
    def set_adapter(self, adapter: MetabaseApi) -> None:
        """Sets the adapter on an object

        Args:
            adapter (MetabaseApi): Connection to MetabaseApi
        """
        self._adapter = adapter
        self._server_version = adapter.server_version

    @log_call
    def refresh(self: T) -> T:
        """Returns new copy of item from API

        Returns:
            T: self
        """
        if self._adapter and self._adapter.server_version >= Version("v0.40"):
            result = self._adapter.get(endpoint=self._BASE_EP.format(id=self.id))
            if isinstance(result, dict):
                obj = self.__class__(**result)
                obj.set_adapter(adapter=self._adapter)
                return obj
        elif self._adapter:
            # In version 0.39 less information is returned when using specific ids
            result = self._adapter.get(endpoint=self._BASE_EP.replace("{id}", ""))
            for item in result:
                if isinstance(item, dict) and self.id == item["id"]:
                    obj = self.__class__(**item)
                    obj.set_adapter(adapter=self._adapter)
                    return obj
        return self

    def _make_update(self: T, **kwargs: Any) -> T:
        """Generic method for updating an object

        Args:
            payloads (dict): Details of update

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            T: Object of the relevant type
        """
        if self._adapter:
            # Eliminate parameters that were not provided
            changes = {
                k: v for k, v in kwargs.items() if not isinstance(v, MissingParam)
            }

            # Determine if there are any differences to update
            current_definition = self.dict()
            changes = {
                k: v
                for k, v in changes.items()
                if k in current_definition and v != current_definition[k]
            }

            if len(changes) == 0:
                return self

            result = self._adapter.put(
                endpoint=self._BASE_EP.format(id=self.id), json=changes
            )
            if isinstance(result, dict):
                obj = self.__class__(**result)
                obj.set_adapter(adapter=self._adapter)
                return obj
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def archive(self: T) -> T:
        """Generic method for archiving an object

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            T: Object of the relevant type
        """
        payload = {"id": self.id, "archived": True}
        if self._adapter:
            result = self._adapter.put(
                endpoint=self._BASE_EP.format(id=self.id), json=payload
            )
            if isinstance(result, dict):
                obj = self.__class__(**result)
                obj.set_adapter(self._adapter)
                return obj
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def unarchive(self: T) -> T:
        """Generic method for unarchiving an object

        Raises:
            InvalidParameters: Targets and jsons are both None

        Returns:
            T: Object of the relevant type
        """
        payload = {"id": self.id, "archived": False}
        if self._adapter:
            result = self._adapter.put(
                endpoint=self._BASE_EP.format(id=self.id), json=payload
            )
            if isinstance(result, dict):
                obj = self.__class__(**result)
                obj.set_adapter(self._adapter)
                return obj
            raise TypeError(f"Expected dict, received {type(result)}")
        raise AttributeError("Adapter not set on object")

    @log_call
    def delete(self) -> None:
        """Method to delete an object

        Raises:
            InvalidParameters: Invalid target
        """
        if self._adapter:
            _ = self._adapter.delete(endpoint=self._BASE_EP.format(id=self.id))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['metabase_tools',
 'metabase_tools.endpoints',
 'metabase_tools.models',
 'metabase_tools.tools',
 'metabase_tools.utils']

package_data = \
{'': ['*']}

install_requires = \
['packaging>=21.3,<22.0',
 'pydantic>=1.9.1,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'types-requests>=2.28.9,<3.0.0']

setup_kwargs = {
    'name': 'metabase-tools',
    'version': '0.12.0',
    'description': 'Unofficial API wrapper for Metabase plus additional helper tools',
    'long_description': '# metabase-tools for python\n\nUnofficial API wrapper for Metabase plus additional helper tools\n\n[Docs](https://j01101111sh.github.io/metabase-tools/)\n',
    'author': 'Josh Odell',
    'author_email': 'j01101111sh@gmail.com',
    'maintainer': 'Josh Odell',
    'maintainer_email': 'j01101111sh@gmail.com',
    'url': 'https://j01101111sh.github.io/metabase-tools/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Classes related to server settings for Metabase"""

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel, Field, PrivateAttr

from metabase_tools.exceptions import MetabaseApiException
from metabase_tools.utils.logging_utils import log_call

if TYPE_CHECKING:
    from metabase_tools.metabase import MetabaseApi

logger = getLogger(__name__)


def replace_hyphens(python_member_name: str) -> str:
    """Used as an alias generator to convert Metabase settings names to Python-friendl\
        y names

    Args:
        python_member_name (str): Python friendly name

    Returns:
        str: Metabase setting name
    """
    return python_member_name.replace("_", "-")


class Setting(BaseModel):
    """Individual setting on the server"""

    _adapter: Optional[MetabaseApi] = PrivateAttr(None)

    key: str
    value: Optional[Any]
    is_env_setting: bool
    env_name: str
    description: str
    default: Optional[Any]

    @log_call
    def set_adapter(self, adapter: MetabaseApi) -> None:
        """Sets the adapter on an object

        Args:
            adapter (MetabaseApi): Connection to MetabaseApi
        """
        self._adapter = adapter

    @log_call
    def update(self, new_value: Any) -> dict[str, Any]:
        """Updates the setting to the provided value after testing basic compatibility

        Args:
            new_value (Any): Desired value for the setting

        Raises:
            TypeError: new_value was not the same type as current value
            MetabaseApiException

        Returns:
            dict[str, Any]: Status of setting change
        """
        if not self._value_type_compatible(new_value):
            raise TypeError
        if self._adapter:
            result = self._adapter.put(
                endpoint=f"/setting/{self.key}", json={"value": new_value}
            )
            if isinstance(result, dict):
                new_def = {
                    setting["key"]: setting
                    for setting in self._adapter.get(endpoint="/setting")
                    if isinstance(setting, dict)
                }[self.key]
                if isinstance(new_def, dict):
                    self.value = new_def["value"]
                return result
        raise MetabaseApiException

    def _value_type_compatible(self, new_value: Any) -> bool:
        if self.value is None or new_value is None:
            return True
        if isinstance(new_value, type(self.value)):
            return True
        return False


class ServerSettings(BaseModel, alias_generator=replace_hyphens, extra="ignore"):
    """Settings for a Metabase server"""

    _adapter: MetabaseApi = PrivateAttr(None)

    admin_email: Setting
    anon_tracking_enabled: Setting
    application_colors: Setting
    application_favicon_url: Setting
    application_logo_url: Setting
    application_name: Setting
    available_locales: Setting
    available_timezones: Setting
    breakout_bin_width: Setting
    breakout_bins_num: Setting
    check_for_updates: Setting
    custom_formatting: Setting
    custom_geojson: Setting
    email_configured: Setting = Field(..., alias="email-configured?")
    email_from_address: Setting
    email_smtp_host: Setting
    email_smtp_password: Setting
    email_smtp_port: Setting
    email_smtp_security: Setting
    email_smtp_username: Setting
    embedding_app_origin: Setting
    embedding_secret_key: Setting
    enable_audit_app: Setting = Field(..., alias="enable-audit-app?")
    enable_embedding: Setting
    enable_enhancements: Setting = Field(..., alias="enable-enhancements?")
    enable_nested_queries: Setting
    enable_password_login: Setting
    enable_public_sharing: Setting
    enable_query_caching: Setting
    enable_sandboxes: Setting = Field(..., alias="enable-sandboxes?")
    enable_sso: Setting = Field(..., alias="enable-sso?")
    enable_whitelabeling: Setting = Field(..., alias="enable-whitelabeling?")
    enable_xrays: Setting
    engines: Setting
    field_filter_operators_enabled: Optional[Setting] = Field(
        alias="field-filter-operators-enabled?"
    )
    ga_code: Setting
    google_auth_auto_create_accounts_domain: Setting
    google_auth_client_id: Setting
    has_sample_dataset: Optional[Setting] = Field(alias="has-sample-dataset?")
    hide_embed_branding: Setting = Field(..., alias="hide-embed-branding?")
    humanization_strategy: Setting
    landing_page: Setting
    ldap_attribute_email: Setting
    ldap_attribute_firstname: Setting
    ldap_attribute_lastname: Setting
    ldap_bind_dn: Setting
    ldap_configured: Setting = Field(..., alias="ldap-configured?")
    ldap_enabled: Setting
    ldap_group_base: Setting
    ldap_group_mappings: Setting
    ldap_group_sync: Setting
    ldap_host: Setting
    ldap_password: Setting
    ldap_port: Setting
    ldap_security: Setting
    ldap_user_base: Setting
    ldap_user_filter: Setting
    map_tile_server_url: Setting
    metabot_enabled: Optional[Setting]
    password_complexity: Setting
    premium_embedding_token: Setting
    premium_features: Optional[Setting]
    query_caching_max_kb: Setting
    query_caching_max_ttl: Setting
    query_caching_min_ttl: Setting
    query_caching_ttl_ratio: Setting
    redirect_all_requests_to_https: Setting
    redshift_fetch_size: Optional[Setting]
    report_timezone: Setting
    report_timezone_short: Setting
    search_typeahead_enabled: Setting
    setup_token: Setting
    show_homepage_data: Setting
    show_homepage_xrays: Setting
    site_locale: Setting
    site_name: Setting
    site_url: Setting
    slack_token: Setting
    source_address_header: Setting
    ssh_heartbeat_interval_sec: Setting
    ssl_certificate_public_key: Setting
    start_of_week: Setting
    version: Setting
    version_info: Setting
    version_info_last_checked: Setting

    @log_call
    def set_adapter(self, adapter: MetabaseApi) -> None:
        """Sets the adapter on an object

        Args:
            adapter (MetabaseApi): Connection to MetabaseApi
        """
        self._adapter = adapter

        for setting in self.__dict__.values():
            if isinstance(setting, Setting):
                setting.set_adapter(adapter)



"""Classes related to card endpoints
"""

from __future__ import annotations

from logging import getLogger
from typing import Any, ClassVar, Optional

from metabase_tools.endpoints.generic_endpoint import Endpoint
from metabase_tools.models.card_model import CardItem
from metabase_tools.models.generic_model import MissingParam
from metabase_tools.utils.logging_utils import log_call

logger = getLogger(__name__)


class Cards(Endpoint[CardItem]):
    """Card related endpoint methods"""

    _BASE_EP: ClassVar[str] = "/card"
    _STD_OBJ: ClassVar[type] = CardItem

    _required_params: ClassVar[list[str]] = [
        "visualization_settings",
        "name",
        "dataset_query",
        "display",
    ]

    @log_call
    def get(self, targets: Optional[list[int]] = None) -> list[CardItem]:
        """Fetch list of cards

        Args:
            targets (list[int], optional): If provided, the list of cards to fetch

        Returns:
            list[CardItem]
        """
        return super().get(targets=targets)

    def _make_create(self, **kwargs: Any) -> CardItem:
        """Makes create request

        Args:
            self (CardItem)

        Returns:
            CardItem: self
        """
        return super()._make_create(**kwargs)

    @log_call
    def create(
        self,
        visualization_settings: dict[str, Any] | MissingParam = MissingParam(),
        name: str | MissingParam = MissingParam(),
        dataset_query: dict[str, Any] | MissingParam = MissingParam(),
        display: str | MissingParam = MissingParam(),
        description: Optional[str | MissingParam] = MissingParam(),
        collection_position: Optional[int | MissingParam] = MissingParam(),
        result_metadata: Optional[list[dict[str, Any]] | MissingParam] = MissingParam(),
        metadata_checksum: Optional[str | MissingParam] = MissingParam(),
        collection_id: Optional[int | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> CardItem:
        """Creates a new card

        Args:
            visualization_settings (dict[str, Any])
            name (str)
            dataset_query (dict[str, Any])
            display (str)
            description (str, optional)
            collection_position (int, optional)
            result_metadata (list[dict[str, Any]], optional)
            metadata_checksum (str, optional)
            collection_id (int, optional)

        Returns:
            CardItem
        """
        return self._make_create(
            visualization_settings=visualization_settings,
            description=description,
            collection_position=collection_position,
            result_metadata=result_metadata,
            metadata_checksum=metadata_checksum,
            collection_id=collection_id,
            name=name,
            dataset_query=dataset_query,
            display=display,
            **kwargs,
        )

    @log_call
    def search(
        self,
        search_params: list[dict[str, Any]],
        search_list: Optional[list[CardItem]] = None,
    ) -> list[CardItem]:
        """Method to search a list of cards meeting a list of parameters

        Args:
            search_params (list[dict]): Each dict contains search criteria and returns\
                 1 result
            search_list (list[CardItem], optional): Provide to search an existing \
                list, by default pulls from API

        Returns:
            list[CardItem]: List of cards of the relevant type
        """
        return super().search(search_params=search_params, search_list=search_list)

    @log_call
    def embeddable(self) -> list[CardItem]:
        """Fetch list of cards with embedding enabled

        Raises:
            EmptyDataReceived: If no cards have embedding enabled

        Returns:
            list[CardItem]: List of cards with embedding enabled
        """
        cards = self._adapter.get(endpoint="/card/embeddable")
        card_ids = [card["id"] for card in cards if isinstance(card, dict)]
        result = self._adapter.cards.get(card_ids)
        return result



"""Classes related to collection endpoints
"""

from __future__ import annotations

from logging import getLogger
from typing import Any, ClassVar, Optional

from metabase_tools.endpoints.generic_endpoint import Endpoint
from metabase_tools.models.collection_model import CollectionItem
from metabase_tools.models.generic_model import MissingParam
from metabase_tools.utils.logging_utils import log_call

logger = getLogger(__name__)


class Collections(Endpoint[CollectionItem]):
    """Card related endpoint methods"""

    _BASE_EP: ClassVar[str] = "/collection"
    _STD_OBJ: ClassVar[type] = CollectionItem
    _required_params: ClassVar[list[str]] = ["name", "color"]

    @log_call
    def get(self, targets: Optional[list[int]] = None) -> list[CollectionItem]:
        """Fetch list of collections

        Args:
            targets (list[int], optional): If provided, the list of collections to fetch

        Returns:
            list[CollectionItem]
        """
        return super().get(targets=targets)

    def _make_create(self, **kwargs: Any) -> CollectionItem:
        """Makes create request

        Args:
            self (CollectionItem)

        Returns:
            CollectionItem: self
        """
        return super()._make_create(**kwargs)

    @log_call
    def create(
        self,
        name: str | MissingParam = MissingParam(),
        color: str | MissingParam = MissingParam(),
        description: Optional[str | MissingParam] = MissingParam(),
        parent_id: Optional[int | MissingParam] = MissingParam(),
        namespace: Optional[str | MissingParam] = MissingParam(),
        **kwargs: Any,
    ) -> CollectionItem:
        """Creates a new collection

        Args:
            name (str, optional)
            color (str, optional)
            description (str, optional)
            parent_id (int, optional)
            namespace (str, optional)

        Returns:
            CollectionItem
        """
        return self._make_create(
            name=name,
            color=color,
            description=description,
            parent_id=parent_id,
            namespace=namespace,
            **kwargs,
        )

    @log_call
    def search(
        self,
        search_params: list[dict[str, Any]],
        search_list: Optional[list[CollectionItem]] = None,
    ) -> list[CollectionItem]:
        """Method to search a list of cards meeting a list of parameters

        Args:
            search_params (list[dict]): Each dict contains search criteria and returns\
                 1 result
            search_list (list[CollectionItem], optional): Provide to search an existing\
                 list, by default pulls from API

        Returns:
            list[CollectionItem]: List of cards of the relevant type
        """
        return super().search(search_params=search_params, search_list=search_list)

    @log_call
    def get_tree(self) -> list[dict[str, Any]]:
        """Collection tree

        Args:
            adapter (MetabaseApi): Connection to Metabase API

        Raises:
            EmptyDataReceived: No data returned from API

        Returns:
            list[dict]: Representation of collection tree
        """
        result = self._adapter.get(endpoint="/collection/tree")
        if isinstance(result, list) and all(
            isinstance(record, dict) for record in result
        ):
            return result
        raise TypeError(f"Expected list[dict], received {type(result)}")

    @staticmethod
    def _flatten_tree(parent: dict[str, Any], path: str = "/") -> list[dict[str, Any]]:
        """Recursive function to flatten collection tree to show the full path for all\
             collections

        Args:
            parent (dict): Parent collection
            path (str, optional): Path to parent collection. Defaults to "/".

        Returns:
            list[dict]: Flattened list
        """
        children = []
        for child in parent["children"]:
            children.append(
                {
                    "id": child["id"],
                    "name": child["name"],
                    "path": f'{path}/{parent["name"]}/{child["name"]}'.replace(
                        "//", "/"
                    ),
                }
            )
            if "children" in child and len(child["children"]) > 0:
                grandchildren = Collections._flatten_tree(
                    child, f'{path}/{parent["name"]}'.replace("//", "/")
                )
                if isinstance(grandchildren, list):
                    children.extend(grandchildren)
                else:
                    children.append(grandchildren)
        return children

    @log_call
    def get_flat_list(self) -> list[dict[str, Any]]:
        """Flattens collection tree so the full path of each collection is shown

        Args:
            adapter (MetabaseApi): Connection to Metabase API

        Returns:
            list[dict]: Flattened collection tree
        """
        tree = self.get_tree()
        folders = []
        for root_folder in tree:
            if root_folder["personal_owner_id"] is not None:  # Skips personal folders
                continue
            folders.append(
                {
                    "id": root_folder["id"],
                    "name": root_folder["name"],
                    "path": f'/{root_folder["name"]}',
                }
            )
            folders.extend(Collections._flatten_tree(root_folder))
        return folders

    @log_call
    def graph(self) -> dict[str, Any]:
        """Graph of collection permissions

        Returns:
            dict: graph of collection
        """
        result = self._adapter.get(endpoint="/collection/graph")
        if isinstance(result, dict):
            return result
        raise TypeError(f"Expected dict, received {type(result)}")
